(* solyga@gmx.de, 2019-03-10 *)
IMPLEMENTATION MODULE Clock;

FROM SYSTEM IMPORT CAST;
IMPORT IO;

VAR
  daysUntil: ARRAY (* leapYear *) BOOLEAN, TMonth OF CARDINAL;


(***** date *****)

PROCEDURE IsLeapYear( y: CARDINAL ): BOOLEAN;
BEGIN
  (* average = 97/400 *)
  RETURN ( y MOD 4 = 0 ) & ( y MOD 100 # 0 ) OR ( y MOD 400 = 0 );
END IsLeapYear;

PROCEDURE DaysInYear( y: CARDINAL ): CARDINAL;
BEGIN
  (* average 365.2425 *)
  RETURN 365 + ORD( IsLeapYear( y ) );
END DaysInYear;

(* crashes only for p = 0 OR ( p = 1 ) & ( w = MAX( CARDINAL ) ) *)
PROCEDURE MultiplesBetween( a, b, p: CARDINAL ): CARDINAL;
VAR
  w, d: CARDINAL;
BEGIN
  IF a >= b THEN RETURN 0 END;
  w := b - a;                    (* 0 would be ok *)
  d := a MOD p;
  IF d # 0 THEN d := p - d END;  (* dist between a and next multiple of p MOD p *)
  RETURN w DIV p + ORD( w MOD p > d );
END MultiplesBetween;

(* full range *)
PROCEDURE LeapYearsBetweenYears( y0, y1: CARDINAL ): CARDINAL;
BEGIN
  RETURN MultiplesBetween( y0, y1,   4 ) -
         MultiplesBetween( y0, y1, 100 ) +
         MultiplesBetween( y0, y1, 400 );
END LeapYearsBetweenYears;

PROCEDURE DaysBetweenYears( y0, y1: CARDINAL ): CARDINAL;
BEGIN
  IF y0 >= y1 THEN RETURN 0 END;
  RETURN 365 * ( y1 - y0 ) + LeapYearsBetweenYears( y0, y1 );
END DaysBetweenYears;

PROCEDURE IsShortMonth( m: TMonth ): BOOLEAN;
BEGIN
  RETURN ODD( ( m - 1 ) MOD 7 );
END IsShortMonth;

PROCEDURE DaysInMonth( m: TMonth; leapYear: BOOLEAN ): CARDINAL;
BEGIN
  IF m = 2 THEN
    RETURN 28 + ORD( leapYear );
  ELSE
    RETURN 31 - ( ( m - 1 ) MOD 7 ) MOD 2;
  END;
END DaysInMonth;

PROCEDURE DaysInYearUntil( m: TMonth; leapYear: BOOLEAN ): CARDINAL;
BEGIN
  RETURN daysUntil[ leapYear, m ];
END DaysInYearUntil;

PROCEDURE IsValidDate( y, m, d: CARDINAL ): BOOLEAN;
VAR
  l: BOOLEAN;
BEGIN
  l := IsLeapYear( y );
  RETURN ( minYear <= y ) & ( y < maxYear ) &
         ( 0 < m ) & ( m <= maxMonth ) &
         ( 0 < d ) & ( d <= DaysInMonth( m, l ) )
         OR
         ( y = maxYear ) & (
           ( 0 < m ) & ( m < maxDateMonth ) &
           ( 0 < d ) & ( d <= DaysInMonth( m, l ) )
           OR
           ( m = maxDateMonth ) &
           ( 0 < d ) & ( d <= maxDateDay )
         );
END IsValidDate;

PROCEDURE PackDate( y, m, d: CARDINAL ): TDate;
BEGIN
  IF ~ IsValidDate( y, m, d ) THEN RETURN 0 END;
  RETURN DaysBetweenYears( minYear, y ) + DaysInYearUntil( m, IsLeapYear(y) ) + ( d - 1 );
END PackDate;

PROCEDURE FindMonth( x: CARDINAL; leapYear: BOOLEAN ): TMonth;
VAR
  a, b, m: CARDINAL;
BEGIN
  a := minMonth;
  b := maxMonth + 1;
  WHILE a < b DO
    m := a + ( b - a ) DIV 2;
    IF daysUntil[ leapYear, m ] <= x THEN a := m + 1 ELSE b := m END;
  END;
  RETURN a - 1;
END FindMonth;

PROCEDURE UnpackDate( x: TDate; VAR y, m, d: CARDINAL );
VAR
  n: CARDINAL;
  l: BOOLEAN;
BEGIN
  y := minYear + x DIV 365;
  n := DaysBetweenYears( minYear, y );
  WHILE n > x DO
    DEC( y );
    n := DaysBetweenYears( minYear, y );
  END;
  DEC( x, n );
  l := IsLeapYear( y );
  m := FindMonth( x, l );
  d := x - DaysInYearUntil( m, l ) + 1;
END UnpackDate;

PROCEDURE WeekDay( d: TDate ): TWeekDay;
CONST
  wd0 = minDateWeekDay;
BEGIN
  RETURN TWeekDay( ( ORD( wd0 ) + d ) MOD 7 );
END WeekDay;

(* following ISO 8601: year has 53 cws  <=>  year-01-01 is Thursday OR year-12-31 is Thursday *)
PROCEDURE CalWeeks( y: CARDINAL ): CARDINAL;
BEGIN
  IF ( WeekDay( PackDate(y,1,1) ) = wdThu ) OR ( WeekDay( PackDate(y,12,31) ) = wdThu ) THEN RETURN 53 END;
  RETURN 52;
END CalWeeks;

(* following ISO 8601: cw 1  :=  week that contains the 4th of January *)
PROCEDURE CalWeek( x: TDate ): TCalWeek;
VAR
  y, n, m: CARDINAL;
BEGIN
  y := minYear + x DIV 365;
  n := DaysBetweenYears( minYear, y );
  WHILE n > x DO
    DEC( y );
    n := DaysBetweenYears( minYear, y );
  END;
  INC( n, 3 );                 (* n = PackDate( y, 1, 4 ) *)
  DEC( n, ORD( WeekDay(n) ) ); (* n is start date of cw 1 of year y *)
  IF x < n THEN RETURN CalWeeks( y - 1 ) END;
  m := 1 + ( x - n ) DIV 7;
  IF m <= CalWeeks( y ) THEN RETURN m END;
  RETURN 1;
END CalWeek;

(***** time *****)


PROCEDURE IsValidTime( h, m, s, f: CARDINAL ): BOOLEAN;
BEGIN
  RETURN ( h < 24 ) & ( m < 60 ) & ( s < 60 ) & ( f < 1000 );
END IsValidTime;

PROCEDURE PackTime( h, m, s, f: CARDINAL ): TTime;
BEGIN
  IF ~ IsValidTime( h, m, s, f ) THEN RETURN 0 END;
  RETURN ( ( ( h * 60 ) + m ) * 60 + s ) * 1000 + f;
END PackTime;

PROCEDURE UnpackTime( t: TTime; VAR h, m, s, f: CARDINAL );
BEGIN
  IF t > maxTime THEN t := 0 END;
  f := t MOD 1000;
  t := t DIV 1000;
  s := t MOD 60;
  t := t DIV 60;
  m := t MOD 60;
  h := t DIV 60;
END UnpackTime;


(***** tests *****)


PROCEDURE DaysBetweenYears2( y0, y1: CARDINAL ): CARDINAL;
VAR
  y, x: CARDINAL;
BEGIN
  IF y0 >= y1 THEN RETURN 0 END;
  x := 0;
  FOR y := y0 TO y1 - 1 DO
    INC( x, DaysInYear( y ) );
  END;
  RETURN x;
END DaysBetweenYears2;

PROCEDURE TestDaysBetweenYears();
CONST
  min = 1000;
  max = 3000;
VAR
  y0, y1, d, d2: CARDINAL;
BEGIN
  IO.WrStr( 'Clock.TestDaysBetweenYears ...' );
  IO.WrLn();
  FOR y0 := min TO max DO
    FOR y1 := min TO max DO
      d := DaysBetweenYears( y0, y1 );
      d2 := DaysBetweenYears2( y0, y1 );
      IF d # d2 THEN
        IO.WrCard( y0, 7 );
        IO.WrCard( y1, 7 );
        IO.WrCard( d, 7 );
        IO.WrCard( d2, 7 );
        IO.WrLn();
      END;
    END;
  END;
  IO.WrStr( 'Clock.TestDaysBetweenYears done.' );
  IO.WrLn();
END TestDaysBetweenYears;

PROCEDURE DaysInYearUntil2( m: TMonth; leapYear: BOOLEAN ): CARDINAL;
VAR
  i, x: CARDINAL;
BEGIN
  x := 0;
  FOR i := 1 TO m - 1 DO
    INC( x, DaysInMonth( i, leapYear ) );
  END;
  RETURN x;
END DaysInYearUntil2;

PROCEDURE DaysInYearUntil3( m: TMonth; leapYear: BOOLEAN ): CARDINAL;
CONST
  daysUntilAug = 4 * 31 + 3 * 30;
VAR
  x, n: CARDINAL;
BEGIN
  n := ( m - 1 ) MOD 7;
  x := n * 30 + ( n + 1 ) DIV 2;
  IF m > 7 THEN INC( x, daysUntilAug ) END;
  IF m > 2 THEN DEC( x, 2 - ORD( leapYear ) ) END;
  RETURN x;
END DaysInYearUntil3;

PROCEDURE TestDaysInYearUntil();
VAR
  y: CARDINAL;
  m: TMonth;
  l: BOOLEAN;
  a, b, c: CARDINAL;
BEGIN
  IO.WrStr( 'Clock.TestDaysInYearUntil ...' );
  IO.WrLn();
  FOR y := 1000 TO 3000 DO
    l := IsLeapYear( y );
    FOR m := minMonth TO maxMonth DO
      a := DaysInYearUntil2( m, l );
      b := DaysInYearUntil3( m, l );
      c := DaysInYearUntil( m, l );
      IF ( a # b ) OR ( b # c ) THEN
        IO.WrCard( y, 4 );
        IO.WrCard( m, 3 );
        IO.WrCard( a, 4 );
        IO.WrCard( b, 4 );
        IO.WrCard( c, 4 );
        IO.WrLn();
      END;
    END;
  END;
  IO.WrStr( 'Clock.TestDaysInYearUntil done.' );
  IO.WrLn();
END TestDaysInYearUntil;

PROCEDURE ShowDaysInYearUntil();
VAR
  y: TYear;
  m: TMonth;
  x: CARDINAL;
BEGIN
  y := minYear;
  WHILE IsLeapYear( y ) DO INC( y ) END;
  IO.WrLn();
  IO.WrStr( 'TYPE' );
  IO.WrLn();
  IO.WrStr( '  TDaysUntil = ARRAY TMonth OF CARDINAL;' );
  IO.WrLn();
  IO.WrStr( 'CONST' );
  IO.WrLn();
  IO.WrStr( '  daysUntilN = TDaysUntil{ ' );
  FOR m := minMonth TO maxMonth DO
    x := DaysInYearUntil2( m, FALSE );
    IO.WrCard( x, 1 );
    IF m < maxMonth THEN
      IO.WrStr( ', ' );
    END;
  END;
  IO.WrStr( ' }; ' );
  IO.WrLn();
  WHILE ~ IsLeapYear( y ) DO INC( y ) END;
  IO.WrStr( '  daysUntilL = TDaysUntil{ ' );
  FOR m := minMonth TO maxMonth DO
    x := DaysInYearUntil2( m, TRUE );
    IO.WrCard( x, 1 );
    IF m < maxMonth THEN
      IO.WrStr( ', ' );
    END;
  END;
  IO.WrStr( ' }; ' );
  IO.WrLn();
  IO.WrLn();
END ShowDaysInYearUntil;

PROCEDURE FindMonth2( x: CARDINAL; leapYear: BOOLEAN ): TMonth;
VAR
  m: TMonth;
BEGIN
  FOR m := maxMonth TO minMonth BY -1 DO
    IF x >= daysUntil[ leapYear, m ] THEN RETURN m END;
  END;
  HALT;
END FindMonth2;

PROCEDURE TestFindMonth();
VAR
  x, a, b: CARDINAL;
  l: BOOLEAN;
BEGIN
  IO.WrStr( 'Clock.TestFindMonth ...' );
  IO.WrLn();
  FOR l := FALSE TO TRUE DO
    FOR x := 0 TO 511 DO
      a := FindMonth2( x, FALSE );
      b := FindMonth( x, FALSE );
      IF a # b THEN
        IO.WrCard( ORD( l ), 1 );
        IO.WrCard( x, 4 );
        IO.WrCard( a, 3 );
        IO.WrCard( b, 3 );
        IO.WrLn();
      END;
    END;
  END;
  IO.WrStr( 'Clock.TestFindMonth done.' );
  IO.WrLn();
END TestFindMonth;

PROCEDURE TestPackDate();
VAR
  y, m, d: CARDINAL;
  x, i: CARDINAL;
BEGIN
  IO.WrStr( 'Clock.TestPackDate ...' );
  IO.WrLn();
  i := 0;
  FOR y := minYear TO maxYear DO
    FOR m := minMonth TO maxMonth DO
      FOR d := minDay TO DaysInMonth( m, IsLeapYear(y) ) DO
        x := PackDate( y, m, d );
        IF x # i THEN
          IO.WrStr( 'no match: ' );
          IO.WrCard( y, 4 );
          IO.WrCard( m, 4 );
          IO.WrCard( d, 4 );
          IO.WrCard( x, 10 );
          IO.WrCard( i, 10 );
          IO.WrLn();
        END;
        IF ~ IsValidDate( y, m, d ) THEN
          IO.WrStr( 'invalid' );
          IO.WrLn();
        END;
        IF i >= MAX( TDate ) THEN
          IO.WrStr( 'Clock.TestPackDate ' );
          IO.WrCard( i, 1 );
          IO.WrStr( ' done' );
          IO.WrLn();
          RETURN;
        END;
        INC( i );
      END;
    END;
  END;
END TestPackDate;

PROCEDURE IsValidDate2( y, m, d: CARDINAL ): BOOLEAN;

  PROCEDURE InRange( y, m, d: CARDINAL ): BOOLEAN;
  BEGIN
    RETURN ( minYear <= y ) & ( y <= maxYear ) &
           ( minMonth <= m ) & ( m <= maxMonth ) &
           ( minDay <= d ) & ( d <= maxDay );
  END InRange;

VAR
  l: BOOLEAN;
BEGIN
  IF ~ InRange( y, m, d ) THEN RETURN FALSE END;
  l := IsLeapYear( y );
  IF d > DaysInMonth( m, l ) THEN RETURN FALSE END;
  IF y < maxDateYear THEN RETURN TRUE END;
  IF m < maxDateMonth THEN RETURN TRUE END;
  IF m > maxDateMonth THEN RETURN FALSE END;
  IF d <= maxDateDay THEN RETURN TRUE END;
  RETURN FALSE;
END IsValidDate2;

PROCEDURE TestIsValidDate();
VAR
  y, m, d: CARDINAL;
  v, v2: BOOLEAN;
  a, b: CARDINAL;
BEGIN
  IO.WrStr( 'Clock.TestIsValidDate ...' );
  IO.WrLn();
  a := 0;
  FOR y := 1000 TO 3000 DO
    FOR m := 0 TO 50 DO
      FOR d := 0 TO 100 DO
        v := IsValidDate( y, m , d );
        v2 := IsValidDate2( y, m, d );
        IF v # v2 THEN
          IO.WrCard( y, 4 );
          IO.WrCard( m, 3 );
          IO.WrCard( d, 3 );
          IO.WrCard( ORD( v ), 2 );
          IO.WrCard( ORD( v2 ), 2 );
          IO.WrLn();
        END;
        IF v THEN
          b := PackDate( y, m, d );
          IF a # b THEN
            IO.WrCard( y, 4 );
            IO.WrCard( m, 3 );
            IO.WrCard( d, 3 );
            IO.WrCard( a, 6 );
            IO.WrCard( b, 6 );
            IO.WrLn();
          END;
          INC( a );
        END;
      END;
    END;
  END;
  IF a # MAX( TDate ) + 1 THEN
    IO.WrStr( 'Range error' );
    IO.WrLn();
  END;
  IO.WrStr( 'Clock.TestIsValidDate done.' );
  IO.WrLn();
END TestIsValidDate;

PROCEDURE CalWeeks2( y: CARDINAL ): CARDINAL;
VAR
  x0, x1: CARDINAL;
BEGIN
  x0 := PackDate( y, 1, 4 );
  DEC( x0, ORD( WeekDay(x0) ) );
  INC( y );
  x1 := PackDate( y, 1, 4 );
  DEC( x1, ORD( WeekDay(x1) ) );
  (* ( x1 - x0 ) MOD 7 = 0 *)
  RETURN ( x1 - x0 ) DIV 7;
END CalWeeks2;

PROCEDURE TestCalWeeks();
VAR
  y, a, b: CARDINAL;
BEGIN
  IO.WrStr( 'Clock.TestCalWeeks ...' );
  IO.WrLn();
  FOR y := minYear TO maxYear - 1 DO
    a := CalWeeks( y );
    b := CalWeeks2( y );
    IF a # b THEN
      IO.WrCard( y, 4 );
      IO.WrCard( a, 3 );
      IO.WrCard( b, 3 );
      IO.WrLn();
    END;
  END;
  IO.WrStr( 'Clock.TestCalWeeks done.' );
  IO.WrLn();
END TestCalWeeks;

PROCEDURE Test();
BEGIN
  TestDaysInYearUntil();
  (*
  TestDaysBetweenYears();
  ShowDaysInYearUntil();
  *)
  TestPackDate();
  TestFindMonth();
  TestIsValidDate();
  TestCalWeeks();
END Test;

PROCEDURE InitDaysInYearUntil();
VAR
  leapYear: BOOLEAN;
  m: TMonth;
  n: CARDINAL;
BEGIN
  FOR leapYear := FALSE TO TRUE DO
    n := 0;
    FOR m := minMonth TO maxMonth DO
      daysUntil[ leapYear, m ] := n;
      INC( n, DaysInMonth( m, leapYear ) );
    END;
  END;
END InitDaysInYearUntil;

BEGIN
  InitDaysInYearUntil();
  (*
  Test();
  *)
END Clock.
